# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Config sorting tests

Note that for the purposes of speed, mods are not removed between tests, and the
removal test occurs at the end. This means that the order of tests may matter,
and you should, when writing tests, assume an arbitrary configuration at the
beginning of the test, and attempt to place the test in such an order as to minimize
the number of changes required to get the desired configuration
"""

import filecmp
import os
import sys
from zipfile import ZipFile

import pytest

from portmod._deprecated.vfs import (
    _cleanup_tmp_archive_dir,
    extract_archive_file_to_tmp,
    sort_vfs,
)
from portmod.config.use import add_global_use, remove_use
from portmod.globals import env
from portmod.loader import load_installed_pkg
from portmod.merge import configure
from portmod.tsort import CycleException
from portmodlib._deprecated import _get_install_dir_dest
from portmodlib._deprecated.vfs import find_file, get_vfs_dirs, list_dir
from portmodlib.archives import list_archive
from portmodlib.atom import Atom

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test-config")
    config = dictionary["config"]
    config_ini = dictionary["config_ini"]
    with open(env.prefix().CONFIG, "w") as configfile:
        print(
            f"""
TEST_CONFIG = r"{config}"
TEST_CONFIG_INI = r"{config_ini}"
""",
            file=configfile,
        )
    yield dictionary
    tear_down_env()


def test_sort_vfs(setup):
    """
    Tests that sorting the config files works properly
    """
    # Install mods
    configure(
        ["test/test-1.0", "test/test2-1.0"],
        update=True,
    )
    pkg1 = load_installed_pkg(Atom("test/test-1.0"))
    pkg2 = load_installed_pkg(Atom("test/test2"))
    assert pkg1 and pkg2
    path1 = os.path.normpath(
        os.path.join(env.prefix().ROOT, _get_install_dir_dest(pkg1))
    )
    path2 = os.path.normpath(
        os.path.join(env.prefix().ROOT, _get_install_dir_dest(pkg2))
    )

    # Check that config is correct
    lines = get_vfs_dirs()
    assert path1 in lines
    assert path2 in lines
    assert lines.index(path1) < lines.index(path2)


def test_user_override(setup):
    """
    Tests that user overrides for vfs sorting work properly
    """
    installpath = os.path.join(env.prefix().CONFIG_DIR, "config", "install.csv")
    os.makedirs(os.path.dirname(installpath), exist_ok=True)

    pkg1 = load_installed_pkg(Atom("test/test"))
    pkg2 = load_installed_pkg(Atom("test/test2"))
    assert pkg1 and pkg2
    path1 = os.path.normpath(
        os.path.join(env.prefix().ROOT, _get_install_dir_dest(pkg1))
    )
    path2 = os.path.normpath(
        os.path.join(env.prefix().ROOT, _get_install_dir_dest(pkg2))
    )

    # Enforce that test overrides test2
    with open(installpath, "w") as file:
        print("test/test, test/test2", file=file)

    configure(
        ["test/test-1.0", "test/test2-1.0"],
        update=True,
    )
    sort_vfs()

    # Check that config is correct
    lines = get_vfs_dirs()
    assert path1 in lines
    assert path2 in lines
    assert lines.index(path1) > lines.index(path2)

    # Enforce that test2 overrides test
    with open(installpath, "w") as file:
        print("test/test2, test/test", file=file)

    sort_vfs()

    # Check that config is correct
    lines = get_vfs_dirs()
    assert path1 in lines
    assert path2 in lines
    assert lines.index(path1) < lines.index(path2)

    os.remove(installpath)


def test_user_cycle(setup):
    """
    Tests that cycles introduced by the user are reported correctly
    """
    installpath = os.path.join(env.prefix().CONFIG_DIR, "config", "install.csv")
    os.makedirs(os.path.dirname(installpath), exist_ok=True)

    # Enforce that test overrides test2
    with open(installpath, "w") as file:
        print("test/test, test/test2", file=file)
        print("test/test2, test/test", file=file)

    try:
        with pytest.raises(CycleException):
            configure(
                ["test/test-1.0", "test/test2-1.0"],
                update=True,
            )
            sort_vfs()
    finally:
        os.remove(installpath)


def test_data_override_flag(setup):
    """
    Tests that mods can conditionally override other mods using DATA_OVERRIDES
    depending on the value of a use flag on the target mod
    """
    # Install mods
    remove_use("foo")
    configure(
        ["test/test6-1.0", "test/test7-1.0"],
        update=True,
    )
    sort_vfs()

    pkg1 = load_installed_pkg(Atom("test/test6"))
    pkg2 = load_installed_pkg(Atom("test/test7"))
    assert pkg1 and pkg2
    path1 = os.path.normpath(
        os.path.join(env.prefix().ROOT, _get_install_dir_dest(pkg1))
    )
    path2 = os.path.normpath(
        os.path.join(env.prefix().ROOT, _get_install_dir_dest(pkg2))
    )

    # Check that config is correct
    lines = get_vfs_dirs()
    assert path1 in lines
    assert path2 in lines
    assert lines.index(path1) < lines.index(path2)

    add_global_use("foo")
    configure(["test/test7-1.0"], update=True)
    sort_vfs()

    lines = get_vfs_dirs()
    assert path1 in lines
    assert path2 in lines
    assert lines.index(path1) > lines.index(path2)


def test_find_file(setup):
    """
    Tests that find_file returns the correct file (last in the vfs order)
    """
    configure(
        ["test/test6-1.0", "test/test7-1.0[foo]"],
        update=True,
    )
    pkg1 = load_installed_pkg(Atom("test/test6"))
    assert pkg1
    assert os.path.abspath(os.path.normpath(find_file("foo.txt"))).startswith(
        os.path.normpath(os.path.join(env.prefix().ROOT, _get_install_dir_dest(pkg1)))
    )
    assert "foo.txt" in list_dir("")


def test_local_vfs(setup):
    """
    Tests that sorting the config files works properly
    """
    # Setup local mod
    test_local_package = os.path.join(env.prefix().LOCAL_MODS, "test_package")
    os.makedirs(test_local_package)

    sort_vfs()

    pkg1 = load_installed_pkg(Atom("test/test-1.0"))
    pkg2 = load_installed_pkg(Atom("test/test2"))
    assert pkg1 and pkg2
    path1 = os.path.normpath(
        os.path.join(env.prefix().ROOT, _get_install_dir_dest(pkg1))
    )
    path2 = os.path.normpath(
        os.path.join(env.prefix().ROOT, _get_install_dir_dest(pkg2))
    )

    # Check that config is correct
    lines = get_vfs_dirs()
    assert path1 in lines
    assert path2 in lines
    assert test_local_package in lines


@pytest.mark.skipif(
    sys.platform == "win32", reason="requires zipinfo command from unzip"
)
def test_archives(setup):
    """
    Tests that list_archive and extract_archive_file perform as expected
    """
    os.chdir(env.TMP_DIR)
    path = "test_file"
    archive_path = os.path.join(env.TMP_DIR, "test_archive.zip")
    with open(path, "w") as file:
        print("foo", file=file)
    with ZipFile(archive_path, "w") as myzip:
        myzip.write(path)

    assert path in list_archive(archive_path)
    extracted_path = extract_archive_file_to_tmp(archive_path, path)
    assert filecmp.cmp(path, extracted_path)
    os.remove(path)
    os.remove(archive_path)
    _cleanup_tmp_archive_dir()


def test_remove_vfs(setup):
    # Remove mods
    configure(["test/test-1.0", "test/test2-1.0"])

    pkg1 = load_installed_pkg(Atom("test/test"))
    pkg2 = load_installed_pkg(Atom("test/test2"))
    assert pkg1 and pkg2
    path1 = os.path.normpath(
        os.path.join(env.prefix().ROOT, _get_install_dir_dest(pkg1))
    )
    path2 = os.path.normpath(
        os.path.join(env.prefix().ROOT, _get_install_dir_dest(pkg2))
    )

    # Remove mods
    configure(["test/test-1.0", "test/test2-1.0"], depclean=True)

    # Check that config is no longer contains their entries
    lines = get_vfs_dirs()
    assert path1 not in lines
    assert path2 not in lines



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import io
import logging
import os
import shutil
import sys
from logging import error, info, warning
from typing import Dict, Iterable, Optional

from portmod.pybuild import Pybuild
from portmodlib.atom import Atom, QualifiedAtom, atom_sat
from portmodlib.colour import bright, green, lblue, lgreen, red, yellow
from portmodlib.fs import onerror
from portmodlib.l10n import l10n
from portmodlib.module_util import _add_redirection

from ._deprecated.rebuild import get_rebuild_manifest
from ._deprecated.vfs import (
    clear_vfs_sort,
    require_vfs_sort,
    sort_vfs,
    vfs_needs_sorting,
)
from ._deps import resolve
from .cfg_protect import get_protected_path, get_redirections, handle_cfg_protect
from .config import get_config
from .config.license import _user_package_accept_license_path
from .config.profiles import get_system
from .config.sets import BUILTIN_SETS, add_set, get_set, remove_set
from .config.use import _user_package_use_path, add_use
from .download import fetchable, find_download, is_downloaded
from .globals import env
from .loader import (
    AmbiguousAtom,
    _sandbox_execute_pybuild,
    load_all_installed,
    load_installed_pkg,
    load_pkg,
    load_pkg_fq,
)
from .modules import (
    clear_module_updates,
    modules_need_updating,
    require_module_updates,
    update_modules,
)
from .package import install_pkg, remove_pkg
from .parsers.flags import add_flag
from .perms import Permissions
from .prompt import prompt_bool
from .query import FlagDesc, LocalFlagDesc, get_flag_desc
from .repo import get_repo
from .repo.keywords import _user_package_accept_keywords_path, add_keyword
from .repo.loader import commit_moved
from .repo.metadata import get_license
from .transactions import (
    Delete,
    New,
    Transactions,
    UseDep,
    generate_transactions,
    print_transactions,
    sort_transactions,
)
from .tsort import CycleException
from .util import KeywordDep, LicenseDep, select_package
from .vdb import VDB


def global_prepare():
    """Performs global system changes which must be done prior to changes"""
    commit_moved()


def global_updates():
    """Performs updates to global configuration"""
    # Fix vfs ordering and update modules

    try:
        sort_vfs()
        clear_vfs_sort()
    except CycleException as e:
        error(f"{e}")

    update_modules()
    clear_module_updates()

    handle_cfg_protect(get_redirections())


def deselect(pkgs: Iterable[str]):
    all_to_remove = []

    for name in pkgs:
        atom = Atom(name)
        to_remove = []
        for mod in get_set("selected-packages"):
            if atom_sat(mod, atom):
                to_remove.append(mod)

        if len(to_remove) == 1:
            info(">>> " + l10n("remove-from-world", atom=green(to_remove[0])))
            all_to_remove.append(to_remove[0])
        elif len(to_remove) > 1:
            raise AmbiguousAtom(atom, to_remove)

    if not all_to_remove:
        print(">>> " + l10n("no-matching-world-atom"))
        return

    if not env.INTERACTIVE or prompt_bool(bright(l10n("remove-from-world-qn"))):
        for mod in all_to_remove:
            remove_set("selected-packages", mod)


def _update_configuration(
    keyword_changes: Iterable[KeywordDep],
    license_changes: Iterable[LicenseDep],
    use_changes: Iterable[UseDep],
    *,
    protect: bool,
):
    def protected_changes_required():
        info("")
        warning(
            l10n(
                "configuration-changes-required",
                command=f"portmod {env.PREFIX_NAME} cfg-update",
            ),
        )
        sys.exit(1)

    def create_protect_file(path: str):
        # get_protected_path will return the same path
        # if the original file does not exist
        if not os.path.exists(path):
            with open(path, "w", encoding="utf-8"):
                pass
        protect_file = get_protected_path(path)
        try:
            shutil.copy2(path, protect_file)
        except FileNotFoundError:
            pass
        _add_redirection(protect_file, path)
        return protect_file

    if keyword_changes:
        for keyword in keyword_changes:
            if keyword.masked:
                error(
                    l10n(
                        "package-masked-keyword",
                        atom=keyword.atom.CPN,
                        arch=env.prefix().ARCH,
                    )
                )
                sys.exit(1)

        print(l10n("necessary-keyword-changes"))
        for keyword in keyword_changes:
            if keyword.keyword.startswith("*"):
                c = red
            else:
                c = yellow
            print("    {} {}".format(green(keyword.atom), c(keyword.keyword)))

        if not env.INTERACTIVE or prompt_bool(l10n("apply-changes-qn")):
            keyword_file = _user_package_accept_keywords_path()
            protect_file = None
            if protect:
                protect_file = create_protect_file(keyword_file)

            for keyword in keyword_changes:
                add_keyword(keyword.atom, keyword.keyword, protect_file=protect_file)
        else:
            sys.exit(1)

    if license_changes:
        warning(l10n("necessary-license-changes"))
        protect_file = None
        for license in license_changes:
            print("    {} {}".format(green(license.atom), license.license))
            # For EULA licenses, display the license and prompt the user to accept
            if license.is_eula and env.INTERACTIVE:
                print()

                def indent(text: str) -> str:
                    return "\n".join(["    " + line for line in text.split("\n")])

                print("    ---", l10n("license-start", license=license.license), "---")
                print(
                    indent(
                        get_license(get_repo(license.repo).location, license.license)
                    )
                )
                print("    ---", l10n("license-end", license=license.license), "---")

            if not env.INTERACTIVE or prompt_bool(l10n("apply-changes-qn")):
                license_file = _user_package_accept_license_path()
                if protect and not protect_file:
                    protect_file = create_protect_file(license_file)

                add_flag(protect_file or license_file, license.atom, license.license)
            else:
                sys.exit(1)

    if use_changes:
        display_flags: Dict[str, FlagDesc] = {}
        for use in use_changes:
            flag = use.flag.lstrip("-")
            desc = get_flag_desc(load_pkg_fq(use.atom), flag)
            if isinstance(desc, LocalFlagDesc):
                display_flags[Atom(desc.pkg.ATOM.CPF).use(flag)] = desc
            else:
                display_flags[flag] = desc or FlagDesc("<missing description>")

        for key, value in display_flags.items():
            print(l10n("use-flag-desc", desc=value, flag=bright(lgreen(key))))

        print()
        print(l10n("necessary-flag-changes"))
        for use in use_changes:
            comment = "\n    ".join(use.comment)
            if comment:
                comment += "\n    "
            if use.flag.startswith("-") and use.oldvalue == use.flag.lstrip("-"):
                print(
                    "   {}{} {} # {}".format(
                        comment,
                        lblue(use.atom.CPF),
                        red(use.flag),
                        l10n("enabled-comment"),
                    )
                )
            elif not use.flag.startswith("-") and use.oldvalue == "-" + use.flag:
                print(
                    "    {}{} {} # {}".format(
                        comment,
                        green(use.atom.CPF),
                        red(use.flag),
                        l10n("disabled-comment"),
                    )
                )
            else:
                print("    {}{} {}".format(comment, green(use.atom.CPF), red(use.flag)))
        if not env.INTERACTIVE or prompt_bool(l10n("apply-changes-qn")):
            protect_file = None
            if protect:
                protect_file = create_protect_file(_user_package_use_path())

            for use in use_changes:
                add_use(
                    use.flag.lstrip("-"),
                    Atom(use.atom.CPF),
                    use.flag.startswith("-"),
                    use.comment,
                    protect_file=protect_file,
                )
        else:
            sys.exit(1)
    if protect and (keyword_changes or use_changes or license_changes):
        protected_changes_required()


def get_messages(pkg: Pybuild):
    """
    Produces messages produced by Pybuild1.warn or Pybuild1.info
    during package phase functions
    """
    warning_path = os.path.join(env.WARNINGS_DIR, pkg.ATOM.CPF)
    info_path = os.path.join(env.WARNINGS_DIR, pkg.ATOM.CPF)
    messages = []
    if os.path.exists(warning_path):
        with open(warning_path, "r") as file:
            messages.append(("warn", file.read()))

    if os.path.exists(info_path):
        with open(info_path, "r") as file:
            messages.append(("info", file.read()))
    return messages


def configure(
    atoms: Iterable[str],
    *,
    delete: bool = False,
    depclean: bool = False,
    auto_depclean: bool = False,
    oneshot: bool = False,
    verbose: bool = False,
    update: bool = False,
    nodeps: bool = False,
    deselect: Optional[bool] = None,
    select: Optional[bool] = None,
    deep: bool = False,
    emptytree: bool = False,
):
    global_prepare()

    # Ensure that we always get the config before performing operations on packages
    # This way the config settings will be available as environment variables.
    get_config()

    targetlist = list(atoms)
    for modstr in targetlist:
        if modstr.startswith("@"):
            # Atom is actually a set. Load set instead
            targetlist.extend(get_set(modstr.replace("@", "")))
            continue

    to_remove = set()
    if delete or depclean:
        for modstr in targetlist:
            if modstr.startswith("@"):
                continue

            skip = False
            atom = Atom(modstr)
            for system_atom in get_system():
                if atom_sat(system_atom, atom):
                    warning(l10n("skipping-system-package", atom=system_atom))
                    skip = True
                    break

            if not skip:
                to_remove.add(atom)

    atomlist = [
        Atom(modstr)
        for modstr in targetlist
        if modstr not in to_remove and not modstr.startswith("@")
    ]

    selected = set(Atom(atom) for atom in atoms if not atom.startswith("@"))
    selected_sets = set(atom[1:] for atom in atoms if atom.startswith("@"))

    explicit = set(selected)
    for selected_set in selected_sets:
        explicit |= get_set(selected_set)

    if delete:
        # Do nothing. We don't care about deps
        transactions = Transactions()
        for atom in to_remove:
            ipkg = load_installed_pkg(atom)
            if not ipkg:
                raise Exception(l10n("not-installed", atom=atom))
            transactions.append(Delete(ipkg))
    elif nodeps:
        fqatoms = []
        enabled_flags = {}
        usedeps = []
        for atom in atomlist:
            pkg, _ = select_package(load_pkg(atom))
            fqatoms.append(pkg.ATOM)
            enabled_flags[pkg.ATOM] = pkg.get_use()
            for flag in atom.USE:
                usedeps.append(
                    UseDep(
                        pkg.ATOM,
                        flag,
                        None,
                        ["# Use requirement passed on command line"],
                    )
                )
                if flag.startswith("-"):
                    enabled_flags[pkg.ATOM].remove(flag.lstrip("-"))
                else:
                    enabled_flags[pkg.ATOM].add(flag)

        selected_cpn = set()
        for atom in selected:
            pkg, _ = select_package(load_pkg(atom))
            selected_cpn.add(QualifiedAtom(pkg.CPN))

        transactions = generate_transactions(
            fqatoms,
            [],
            set() if oneshot or depclean else selected_cpn,
            usedeps,
            enabled_flags,
            update=update,
            emptytree=emptytree,
        )
    else:
        transactions = resolve(
            atomlist,
            to_remove,
            explicit,
            set() if oneshot or depclean else selected,
            set() if oneshot or depclean else selected_sets,
            deep=deep
            or (depclean and not to_remove),  # No argument depclean implies deep
            update=update,
            depclean=auto_depclean or depclean,
            emptytree=emptytree,
        )

    transactions = sort_transactions(transactions)

    # Inform user of changes
    if transactions.pkgs:
        # Don't print transaction list when in quiet mode and no-confirm is passed
        if env.INTERACTIVE or logging.root.level < logging.WARN:
            if delete or depclean:
                print(l10n("to-remove"))
            else:
                print(l10n("to-install"))
            print_transactions(transactions, verbose=verbose)
            print()
    elif (vfs_needs_sorting() or modules_need_updating()) and not transactions.pkgs:
        global_updates()
        info(l10n("nothing-else-to-do"))
        return
    elif not transactions.pkgs:
        info(l10n("nothing-to-do"))
        return

    if transactions.config:
        keyword_changes = list(
            filter(lambda x: isinstance(x, KeywordDep), transactions.config)
        )
        license_changes = list(
            filter(lambda x: isinstance(x, LicenseDep), transactions.config)
        )
        use_changes = list(filter(lambda x: isinstance(x, UseDep), transactions.config))
        _update_configuration(
            keyword_changes,
            license_changes,
            use_changes,
            # Don't protect files when testing.
            protect=not env.INTERACTIVE and not env.TESTING,
        )

    # Create a temporary VDB so that common packages needed for pkg_nofetch and pkg_pretend
    # but that haven't yet been installed can be found by the loader
    if os.path.exists(env.TMP_VDB):
        shutil.rmtree(env.TMP_VDB, onerror=onerror)
    for trans in transactions.pkgs:
        if isinstance(trans, New) and trans.pkg.CATEGORY == "common":
            os.makedirs(os.path.join(env.TMP_VDB, "common", trans.pkg.PN))
            shutil.copy2(
                trans.pkg.FILE, os.path.join(env.TMP_VDB, "common", trans.pkg.PN)
            )

    def get_restricted_fetch(transactions: Transactions):
        # Check for restricted fetch packages and print their nofetch notices
        restricted = 0
        for trans in transactions.pkgs:
            if not isinstance(trans, Delete):
                can_fetch = fetchable(trans.pkg, trans.flags)
                sources = trans.pkg.get_source_manifests(trans.flags)
                to_fetch = [
                    source for source in sources if find_download(source) is None
                ]
                if set(to_fetch) - set(can_fetch) and not is_downloaded(
                    trans.pkg, trans.flags
                ):
                    restricted += len(set(to_fetch) - set(can_fetch))
                    print(
                        bright(yellow(l10n("fetch-instructions", atom=trans.pkg.ATOM)))
                    )
                    state = {
                        "UNFETCHED": to_fetch,
                        "A": sources,
                        "USE": trans.flags,
                    }
                    _sandbox_execute_pybuild(
                        trans.pkg.FILE,
                        "nofetch",
                        Permissions(),
                        init=state,
                    )
                    print()
        return restricted

    restricted_downloads = get_restricted_fetch(transactions)

    tmp_dir = env.TMP_DIR
    # If TMP_DIR doesn't exist, either use the parent, or if that doesn't exist,
    # just create it
    if not os.path.exists(env.TMP_DIR):
        if os.path.exists(os.path.dirname(env.TMP_DIR)):
            tmp_dir = os.path.dirname(env.TMP_DIR)
        else:
            os.makedirs(tmp_dir, exist_ok=True)
    tmp_space = shutil.disk_usage(tmp_dir).free

    pre_merge_warnings = 0
    for trans in transactions.pkgs:
        if not isinstance(trans, Delete) and "pkg_pretend" in trans.pkg.FUNCTIONS:
            info(">>> " + l10n("pkg-pretend", atom=green(trans.pkg.ATOM.CPF)))
            # TODO: There are various variables that should be set on mod during pkg_pretend
            _sandbox_execute_pybuild(
                trans.pkg.FILE,
                "pretend",
                Permissions(),
                save_state=True,
                init={"USE": trans.flags},
            )

            total_size = 0
            for source in trans.pkg.get_source_manifests(trans.flags):
                total_size += source.size

            # We assume that files have a compression ratio of approximately 0.5
            # Thus we want at least twice the size of the archives in free space.
            if total_size * 2 > tmp_space:
                pre_merge_warnings += 1
                warning(
                    l10n(
                        "tmp-space-too-small",
                        dir=env.TMP_DIR,
                        free=tmp_space / 1024 / 1024,
                        size=total_size * 2 / 1024 / 1024,
                    )
                )
            pre_merge_warnings += len(
                {message for typ, message in get_messages(trans.pkg) if typ == "warn"}
            )

    if pre_merge_warnings:
        warning(l10n("pre-merge-warnings", warnings=pre_merge_warnings))
    if restricted_downloads and not env.INTERACTIVE:
        error(l10n("exiting-non-interactive-nofetch"))
        sys.exit(1)
    elif restricted_downloads:
        warning(l10n("restricted-fetch-summary", restricted=restricted_downloads))
        # Loop until all restricted downloads are available
        while restricted_downloads > 0:
            if not prompt_bool(l10n("continue-qn")):
                sys.exit(1)

            new_restricted_downloads = get_restricted_fetch(transactions)
            if new_restricted_downloads == 0:
                break
            if new_restricted_downloads >= restricted_downloads:
                error(l10n("restricted-fetch-unchanged"))
                sys.exit(1)
            restricted_downloads = new_restricted_downloads
    elif env.INTERACTIVE and not prompt_bool(l10n("continue-qn")):
        return

    err = None
    merged = Transactions()
    # Install (or remove) packages in order
    for trans in transactions.pkgs:
        if isinstance(trans, Delete):
            remove_pkg(trans.pkg)
            if deselect is None or deselect:
                if trans.pkg.CPN in get_set("selected-packages"):
                    info(">>> " + l10n("remove-from-world", atom=green(trans.pkg.CPN)))
                    remove_set("selected-packages", trans.pkg.CPN)
            merged.pkgs.append(trans)
        elif install_pkg(trans.pkg, trans.flags):
            if trans.pkg in transactions.new_selected and not oneshot:
                if trans.pkg.CPN not in get_set("selected-packages"):
                    info(">>> " + l10n("add-to-world", atom=green(trans.pkg.CPN)))
                    add_set("selected-packages", trans.pkg.CPN)
            merged.pkgs.append(trans)
        else:
            # Unable to install mod. Aborting installing remaining packages
            err = trans.pkg.ATOM
            break

        require_module_updates()
        if trans.pkg._PYBUILD_VER == 1:
            require_vfs_sort()
            sort_vfs()
            clear_vfs_sort()

    for trans in merged.pkgs:
        messages = get_messages(trans.pkg)

        if messages:
            print()
            print(">>> " + l10n("pkg-messages", atom=bright(green(trans.pkg))))
            for typ, msg in messages:
                if typ == "WARN":
                    warning(msg)
                elif typ == "INFO":
                    info(msg)
            print()

    for set_name in selected_sets:
        if set_name not in BUILTIN_SETS:
            if deselect or delete or depclean:
                remove_set("selected-sets", Atom(set_name))
            else:
                add_set("selected-sets", Atom(set_name))

    transstring = io.StringIO()
    print_transactions(merged, verbose=True, out=transstring, summarize=False)
    if err:
        # There was an error. We report the packages that were successfully merged and
        # note that an error occurred, however we still commit anyway.
        message = (
            l10n("merge-success-and-error", num=len(transactions.pkgs), atom=err)
            + "\n"
            + transstring.getvalue()
        )
    else:
        message = (
            l10n("merge-success", num=len(transactions.pkgs))
            + "\n"
            + transstring.getvalue()
        )

    # Commit changes to installed database
    with VDB(message):
        pass

    # Check if packages were just modified and can be removed from the rebuild set
    # Any transaction type warrants removal, as they were either rebuilt,
    # and thus can be removed, or deleted, and no longer need to be rebuilt
    for atom in get_set("rebuild"):
        installed_pkg = load_installed_pkg(atom)
        if not installed_pkg or installed_pkg.CPN in [
            trans.pkg.CPN for trans in merged.pkgs
        ]:
            remove_set("rebuild", atom)

    for pkg in load_all_installed():
        # If no pybuild1 packages are installed, the rebuild feature isn't needed
        if pkg._PYBUILD_VER == 1:
            info(l10n("checking-rebuild"))
            break

    # Check if packages need to be added to rebuild set
    for pkg in load_all_installed():
        if pkg.CPN not in get_set("rebuild") and pkg.INSTALLED_REBUILD_FILES:
            seen = set()
            for entry in get_rebuild_manifest(
                pkg.get_installed_env().get("REBUILD_FILES", [])
            ):
                seen.add(entry.name)
                if pkg.INSTALLED_REBUILD_FILES.get(entry.name) != entry:
                    add_set("rebuild", pkg.CPN)
                    break
            if not all(entry in seen for entry in pkg.INSTALLED_REBUILD_FILES.entries):
                add_set("rebuild", pkg.CPN)

    if get_set("rebuild"):
        warning(l10n("rebuild-message"))
        for atom in get_set("rebuild"):
            print("    {}".format(green(atom)))
        print(
            l10n(
                "rebuild-prompt",
                command=lgreen(f"portmod {env.PREFIX_NAME} merge @rebuild"),
            )
        )

    global_updates()



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Critical News module

See https://www.gentoo.org/glep/glep-0042.html for details

We use yaml instead of txt for ease of parsing, while having almost the same structure
Note that a Body field is used for the article body.
"""

import os
from fnmatch import fnmatch
from glob import glob
from typing import Callable

from portmod.lock import exclusive_lock
from portmodlib.atom import Atom
from portmodlib.colour import blue, bright, green, yellow
from portmodlib.l10n import get_locales, l10n
from portmodlib.parsers.list import read_list, write_list
from portmodlib.portmod import parse_news

from .config import get_config_value, variable_data_dir
from .config.profiles import get_profile_name, profile_exists
from .globals import env
from .loader import load_installed_pkg
from .prefix import get_prefixes
from .repo import Repo


def _get_news_dir():
    return os.path.join(variable_data_dir(), "news")


def get_article_path(repo_root: str, article: str) -> str:
    path = os.path.join(repo_root, "metadata", "news", article)
    for locale in get_locales():
        article_path = os.path.join(path, article + "." + locale + ".yaml")
        if os.path.exists(article_path):
            return article_path
    raise FileNotFoundError(
        f"No news article matching {article} could be found in {repo_root}"
    )


def is_news_visible(repo: Repo, article_name: str) -> bool:
    """
    Returns true if the given news article is visible to the user.
    """
    article = parse_news(get_article_path(repo.location, article_name))

    def get_list(ob, key):
        if isinstance(ob, list):
            li = ob
        elif isinstance(ob, str):
            li = ob.split()
        else:
            raise TypeError(
                f"{key} should contain a string or a list, " f"but contained {ob}"
            )
        return li

    installed_conditions = get_list(
        article.display_if_installed or "", "Display-If-Installed"
    )
    keyword_conditions = get_list(
        article.display_if_keyword or "", "Display-If-Keyword"
    )
    profile_conditions = get_list(
        article.display_if_profile or "", "Display-If-Profile"
    )

    profile_name = get_profile_name()
    if profile_conditions and not any(
        fnmatch(profile_name, profile_condition)
        for profile_condition in profile_conditions
    ):
        return False

    if installed_conditions and not any(
        load_installed_pkg(Atom(atom)) for atom in installed_conditions
    ):
        return False

    ACCEPT_KEYWORDS = set(get_config_value("ACCEPT_KEYWORDS"))
    if keyword_conditions and not any(
        keyword in ACCEPT_KEYWORDS or keyword.lstrip("~") in ACCEPT_KEYWORDS
        for keyword in keyword_conditions
    ):
        return False

    return True


def is_news_unread(repo: Repo, article: str) -> bool:
    """
    Returns true if the given news article needs to be read.
    """
    unread_file = os.path.join(_get_news_dir(), "news-" + repo.name + ".unread")
    if not os.path.exists(unread_file) or article in read_list(unread_file):
        return True

    return False


def mark(repo: Repo, article: str, read=True):
    """Marks the given news article as read"""
    os.makedirs(_get_news_dir(), exist_ok=True)
    if read:
        remove_file = os.path.join(_get_news_dir(), "news-" + repo.name + ".unread")
        add_file = os.path.join(_get_news_dir(), "news-" + repo.name + ".read")
    else:
        add_file = os.path.join(_get_news_dir(), "news-" + repo.name + ".unread")
        remove_file = os.path.join(_get_news_dir(), "news-" + repo.name + ".read")

    if os.path.exists(remove_file):
        contents = read_list(remove_file)

        if article in contents:
            del contents[contents.index(article)]
            write_list(remove_file, contents)
        else:
            return

    with open(add_file, "a") as file:
        print(article, file=file)


def iterate_news(repo: Repo, unread_only=False, visible_only=False):
    news_dir = os.path.join(repo.location, "metadata", "news")
    unread_file = os.path.join(_get_news_dir(), "news-" + repo.name + ".unread")
    if os.path.exists(unread_file):
        unread = set(read_list(unread_file))
    else:
        unread = set()

    if os.path.exists(news_dir):
        for news_item in sorted(os.listdir(news_dir)):
            if (not unread_only or news_item in unread) and (
                not visible_only or is_news_visible(repo, news_item)
            ):
                yield os.path.join(news_dir, news_item)


def display_unread_message():
    total_unread = 0
    for repo in env.prefix().REPOS:
        unread_file = os.path.join(_get_news_dir(), "news-" + repo.name + ".unread")
        if os.path.exists(unread_file):
            unread = len(read_list(unread_file))
            if unread:
                if not total_unread:
                    print()
                print(
                    bright(yellow(" * " + l10n("important")))
                    + l10n("news-unread", unread=unread, repo=repo.name)
                )
            total_unread += unread

    if total_unread:
        print(
            bright(yellow(" * "))
            + l10n(
                "news-read",
                command=bright(green(f"portmod {env.PREFIX_NAME} select news read")),
            )
        )
        print()


def update_news():
    # Update for each prefix
    orig_prefix = env.PREFIX_NAME
    for prefix in get_prefixes():
        env.set_prefix(prefix)
        with exclusive_lock():
            for repo in env.prefix().REPOS:
                os.makedirs(_get_news_dir(), exist_ok=True)
                unread_file = os.path.join(
                    _get_news_dir(), "news-" + repo.name + ".unread"
                )
                skip_file = os.path.join(_get_news_dir(), "news-" + repo.name + ".skip")
                if os.path.exists(skip_file):
                    skip = set(read_list(skip_file))
                else:
                    skip = set()

                repo_news = set()
                for news_dir in iterate_news(repo):
                    news_item = os.path.basename(news_dir)
                    repo_news.add(news_item)
                    if news_item in skip:
                        continue

                    if (
                        profile_exists()
                        and is_news_visible(repo, news_item)
                        and os.path.exists(unread_file)
                    ):
                        with open(unread_file, "a") as file:
                            print(news_item, file=file)

                    with open(skip_file, "a+") as file:
                        print(news_item, file=file)

                if not os.path.exists(unread_file):
                    # If unread file doesn't exist, this must be a newly added repository
                    # In this case, all news is old news, so it gets added to the skip file
                    # above, but not to the unread file
                    with open(unread_file, "w") as file:
                        file.write("")
                else:
                    # Prune any stale entries from the unread file
                    contents = read_list(unread_file)
                    new_contents = [line for line in contents if line in repo_news]
                    if contents != new_contents:
                        write_list(unread_file, new_contents)

    env.set_prefix(orig_prefix)
    return True


def read_news(index=None, unread_only=False):
    i = 0
    for repo in env.prefix().REPOS:
        for news_dir in iterate_news(repo, unread_only=unread_only):
            if index is not None and i < index:
                i += 1
                continue
            elif index is not None and i > index:
                break

            news_item = os.path.basename(news_dir)
            article = parse_news(get_article_path(repo.location, news_item))

            max_len = max(
                len(article.title),
                len(article.author),
                len(article.translator or ""),
                len(article.posted),
                len(article.revision),
            )
            print(bright(green((news_item))))
            print("  " + l10n("title").ljust(max_len) + article.title)
            print("  " + l10n("author").ljust(max_len) + article.author)
            if article.translator:
                print("  " + l10n("translator").ljust(max_len) + article.translator)
            print("  " + l10n("posted").ljust(max_len) + article.posted)
            print("  " + l10n("revision").ljust(max_len) + article.revision)
            print()
            print(article.body)
            print()

            mark(repo, news_item)
            i += 1


def add_news_parsers(subparsers, parents):
    news = subparsers.add_parser("news", help=l10n("news-help"), parents=parents)
    news_subparsers = news.add_subparsers()
    news_list = news_subparsers.add_parser(
        "list", help=l10n("news-list-help"), parents=parents
    )
    news_read = news_subparsers.add_parser(
        "read", help=l10n("news-read-help"), parents=parents
    )
    news_read.add_argument(
        "target",
        help=l10n("news-read-target-help"),
        default=l10n("news-read-target-new"),
        metavar="<" + l10n("news-target-placeholder") + ">",
        nargs="?",
    )
    news_unread = news_subparsers.add_parser(
        "unread", help=l10n("news-unread-help"), parents=parents
    )
    news_unread.add_argument(
        "target",
        help=l10n("news-unread-target-help"),
        metavar="<" + l10n("news-target-placeholder") + ">",
    )

    def read_func(args):
        if args.target == l10n("news-read-target-new"):
            read_news(unread_only=True)
        elif args.target == l10n("news-read-target-all"):
            read_news()
        else:
            read_news(int(args.target))

    def list_func(args):
        i = 0
        print(bright(green(l10n("news-items"))))
        for repo in env.prefix().REPOS:
            for path in iterate_news(repo):
                article_name = os.path.basename(path)
                article = parse_news(get_article_path(repo.location, article_name))
                if is_news_unread(repo, article_name):
                    print(
                        f'  {bright("[" + str(i) + "]")}  {blue("N")}  '
                        f"{article.posted}  {blue(article.title)}"
                    )
                else:
                    print(
                        f'  {bright("[" + str(i) + "]")}     '
                        f"{article.posted}  {article.title}"
                    )
                i += 1

    def news_help(args):
        news.print_help()

    def unread_func(args):
        i = 0
        for repo in env.REPOS:
            for path in iterate_news(repo):
                if args.target == "all" or i == int(args.target):
                    article_name = os.path.basename(path)
                    mark(repo, article_name, read=False)
                i += 1

    news.set_defaults(func=news_help)
    news_read.set_defaults(func=read_func)
    news_list.set_defaults(func=list_func)
    news_unread.set_defaults(func=unread_func)


def validate_news(repo_root: str, err: Callable[[str], None]):
    """Validates the news files in the given repository"""
    path = os.path.join(repo_root, "metadata", "news")
    if os.path.exists(path):
        for directory in os.listdir(path):
            files = glob(os.path.join(path, directory, directory + ".*.yaml"))
            for news_file in files:
                article = parse_news(news_file)

                if article.display_if_installed:
                    string = article.display_if_installed
                    for atom in string.split():
                        Atom(atom)
            if not files:
                err(
                    f"News directory {path}/{directory} doesn't contain any files of "
                    f"the form {directory}.*.yaml"
                )



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests the news subsystem
"""

import os
import sys

import pytest

from portmod._cli.main import main
from portmod.merge import configure
from portmod.news import (
    _get_news_dir,
    display_unread_message,
    iterate_news,
    mark,
    read_news,
    update_news,
)
from portmod.repo import get_repo
from portmodlib.parsers.list import read_list

from .env import select_profile, setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=False)
def setup():
    """sets up and tears down test environment"""
    dictionary = setup_env("test")
    yield dictionary
    tear_down_env()


def test_news(setup):
    update_news()

    news_dir = _get_news_dir()
    unread_file = os.path.join(news_dir, "news-test.unread")
    read_file = os.path.join(news_dir, "news-test.read")

    # Check that no news (all news is old news) is unread
    assert not os.path.exists(unread_file) or not read_list(unread_file)

    mark(get_repo("test"), "2020-04-12-test", read=False)
    # Check that article is now unread
    assert "2020-04-12-test" in read_list(unread_file)

    display_unread_message()

    read_news()
    # Check that article is now read
    assert "2020-04-12-test" not in read_list(unread_file)
    assert "2020-04-12-test" in read_list(read_file)


def test_news_installed(setup):
    update_news()

    assert "2020-04-12-installed" not in [
        os.path.basename(article)
        for article in iterate_news(get_repo("test"), visible_only=True)
    ]

    configure(["test/test-1.0"])
    assert "2020-04-12-installed" in [
        os.path.basename(article)
        for article in iterate_news(get_repo("test"), visible_only=True)
    ]

    configure(["test/test-2.0"])
    assert "2020-04-12-installed" not in [
        os.path.basename(article)
        for article in iterate_news(get_repo("test"), visible_only=True)
    ]

    configure(["test/test"], depclean=True)
    assert "2020-04-12-installed" not in [
        os.path.basename(article)
        for article in iterate_news(get_repo("test"), visible_only=True)
    ]


def test_news_profile(setup):
    update_news()

    assert "2020-04-12-profile" not in [
        os.path.basename(article)
        for article in iterate_news(get_repo("test"), visible_only=True)
    ]

    select_profile("test-config")
    assert "2020-04-12-profile" in [
        os.path.basename(article)
        for article in iterate_news(get_repo("test"), visible_only=True)
    ]


def test_news_cli(setup):
    sys.argv = ["portmod", "test", "select", "news", "list"]
    main()

    sys.argv = ["portmod", "test", "select", "news", "unread", "0"]
    main()

    sys.argv = ["portmod", "test", "select", "news", "read"]
    main()



# Copyright 2022 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests the execution environment
"""

import os
import shutil
import sys

import pytest

from portmod.execute import sandbox_execute
from portmod.perms import Permissions


@pytest.mark.parametrize("global_read", [True, False])
def test_symlinks(global_read):
    """Tests that symlinks are handled correctly"""
    try:
        curdir = os.getcwd()
        os.makedirs(os.path.join(curdir, "bar", "baz"), exist_ok=True)
        os.symlink(os.path.join(curdir, "bar", "baz"), "foo")
        if sys.platform == "win32":
            command = ["cmd", "/c", f'copy nul {os.path.join(curdir, "foo", "baf")}']
        else:
            command = ["touch", f"{curdir}/foo/baf"]
        sandbox_execute(
            command,
            Permissions(
                rw_paths=[os.path.join(curdir, "foo")], global_read=global_read
            ),
        )
    finally:
        shutil.rmtree("bar")
        os.remove("foo")



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

# pylint: disable=no-member

import logging
import os
import shutil
from filecmp import cmp
from glob import iglob
from logging import info, warning
from typing import (
    Callable,
    Generator,
    Iterable,
    List,
    Optional,
    Pattern,
    Set,
    Tuple,
    Union,
    cast,
)

from portmodlib._deprecated import File, InstallDir, _get_install_dir_dest, get_files
from portmodlib.atom import Atom
from portmodlib.colour import blue, magenta
from portmodlib.execute import execute
from portmodlib.fs import ci_exists, is_parent, make_unique_filename, patch_dir
from portmodlib.globals import messages_dir, warnings_dir
from portmodlib.l10n import l10n
from portmodlib.pybuild import FullPybuild, get_installed_env
from portmodlib.source import Source
from portmodlib.usestr import check_required_use, use_reduce
from portmodlib.util import fnmatch_list_to_re


def apply_patch(patch: str):
    """
    Applies git patch using Git apply

    Patch files must be in a format that can be applied via git apply. Such patches can
    be produced with ``git diff --no-index ORIG NEW``. The ``--binary`` option can be
    used to produce binary diffs for non-text files.

    Patches must be self-applying.  I.e. they should not rely on paths being passed to
    git apply, and must apply from the default working directory in src_prepare.

    It is recommended that a comment header is included to describe what the patch
    does, where it's from etc.

    args:
        patch: Path to the patch to be applied
    """
    print(l10n("applying-patch", patch=patch))
    execute(["git", "apply", patch, "--stat", "--apply"])


class Pybuild2(FullPybuild):
    """
    The class all Pybuilds should derive.

    The name and path of a pybuild declares the package name, category, version and
    (optionally) revision::

        {CATEGORY}/{PKG_NAME}-{VER}(-r{REV}).pybuild

    Categories and package names may contain lower case letters, numbers and hyphens.
    Versions may contain numbers and dots. Revisions may only contain numbers
    (following the -r prefix). (See the PMS for the complete naming scheme).

    Note that revisions refer to revisions of the pybuild itself, not the package,
    and are used to indicate that the way the mod is configured has changed in a way
    that will impact the installed version. For changes, such as the source files
    moving, that would not impact a mod that is already installed, you do not need to
    update the revision.

    There are certain fields which are defined automatically and may only be available
    in some scopes:

    ============================= ===============================
            Variable                        Scope
    ============================= ===============================
    :py:attr:`Pybuild2.P`         All scopes
    :py:attr:`Pybuild2.PF`        All scopes
    :py:attr:`Pybuild2.PN`        All scopes
    :py:attr:`Pybuild2.CATEGORY`  All scopes
    :py:attr:`Pybuild2.PV`        All scopes
    :py:attr:`Pybuild2.PVR`       All scopes
    :py:attr:`Pybuild2.USE`       All scopes except ``__init__``
    :py:attr:`Pybuild2.WORKDIR`   src_*
    :py:attr:`Pybuild2.T`         All scopes except ``__init__``
    :py:attr:`Pybuild2.D`         :py:func:`Pybuild2.src_install`
    :py:attr:`Pybuild2.FILESDIR`  src_*
    :py:attr:`Pybuild2.ROOT`      src_*, pkg_*
    :py:attr:`Pybuild2.A`         src_*, pkg_nofetch
    :py:attr:`Pybuild2.UNFETCHED` pkg_nofetch
    :py:attr:`Pybuild2.S`         src_*
    ============================= ===============================

    .. [1] Fields which are set automatically and should not be defined in the
           package file. Described in the table above.
    """

    __file__ = __file__

    P: Atom
    """
    The package name and version. [1]_

    E.g.: ``example-suite-1.0``
    """
    PF: Atom
    """
    The package name with version and revision. [1]_

    E.g.: ``example-suite-1.0-r1``
    """
    PN: Atom
    """
    The package name without version. [1]_

    E.g.: ``example-suite``
    """
    CATEGORY: str
    """
    The package's category. [1]_

    E.g. ``base``
    """
    PV: str
    """
    The package's version without revision [1]_

    E.g. ``1.0``
    """
    PR: str
    """
    The package's revision [1]_

    E.g. ``r1``

    Is equal to ``r0`` is no revision is specified
    """
    PVR: str
    """
    The package's version and revision [1]_

    E.g. ``1.0-r1``
    """
    USE: Set[str]
    """
    Enabled use flags [1]_

    Scope: All except __init__
    """
    WORKDIR: str
    """
    The directory where packaging takes place [1]_

    Scope: src_*
    """
    T: str
    """
    Path to a temporary directory which may be used during packaging [1]_

    Scope: All except __init__
    """
    D: str
    """
    The full path of the directory where the package is to be installed. [1]_

    Note that this is a temporary directory and not the final install location.

    Scope: src_install
    """
    ROOT: str
    """
    The full path of the prefix root where packages will be installed

    Note: This functions as both ROOT and SYSROOT (as defined by PMS section 11.1).

    Scope: src_*, pkg_*
    """
    FILESDIR: str
    """
    Path of the directory containing additional repository files

    Scope: src_*
    """
    A: List[Source]
    """
    The list of enabled sources [1]_

    Scope: All except __init__
    """
    UNFETCHED: List[Source]
    """
    The list of sources which need to be fetched [1]_

    Scope: pkg_nofetch
    """
    DEPEND = ""
    """
    Build dependencies.

    The DEPEND field is used to specify packages which need to be installed in order
    for this package to install correctly.

    Most mods do not have build dependencies, however mods that require patching using
    tools external to portmod, or packages that generate content from other sources,
    will need to include their masters, or the other sources, as build dependencies,
    to ensure that they are installed prior to the package being installed.

    Format (both ``DEPEND`` and ``RDEPEND``):
    A list of dependencies in the form of package atoms. All dependencies should
    include both category and package name. Versions should also be included if
    the package depends on a specific version of another mod.
    It is recommended not to include a version number in the dependency unless it is
    known that the package will not work with other versions.

    Ranges of versions can be indicated by prefixing >,<,<=,>= to the atoms.
    E.g. ``>=cat/foo-1.0``

    Specific versions can be indicated by prefixing ``=`` (matches version and
    revision exactly) or ``~`` (matches version, but allows any revision) to the atoms.
    E.g. ``=cat/foo-1.0``

    Use flag dependencies can be specified in the following manner:

    - ``cat/foo[flag]`` - Indicates that flag must be enabled
    - ``cat/foo[flag,flag2]`` - Indicates that both flag and flag2 must be enabled
    - ``cat/foo[!flag]`` - Indicates that flag must be disabled
    - ``cat/foo[flag?]`` - Indicates that flag must be enabled if it is enabled for
      this package
    - ``cat/foo[!flag?]`` - Indicates that flag must be disabled if it is enabled for
      this package

    Atoms can be surrounded by use-conditionals if they are only dependencies when
    that use flag is enabled/disabled.

    E.g. ``flag? ( cat/foo )``

    Atoms can be grouped and prefixed by a ``||`` operator to indicate that any of the
    given packages will satisfy the dependency.

    E.g. ``|| ( cat/foo cat/bar cat/baz )``

    Note that it is required that the parentheses ``(`` ``)`` are separated from the atoms
    by whitespace.

    Packages which cannot be installed at the same time can be marked as blocks using the
    ``!!`` operator. I.e. ``!!cat/foo`` indicates that ``cat/foo`` cannot be installed at
    the same time as the current package.
    """

    RDEPEND = ""
    """
    Runtime dependencies.

    Is used to specify packages which are required at runtime for this package
    to function. The format is the same as for DEPEND
    """
    IUSE = ""
    """
    A field containing a space-separated list of the use flags used by the package.

    IUSE should contain all regular use flags used by this package, both local and
    global. Prefixing the use flags with a + means that the option is enabled by
    default. Otherwise use flags are disabled by default.

    Note that you do not need to include TEXTURE_SIZES type flags in IUSE, but
    USE_EXPAND variables should be included in IUSE.
    """
    KEYWORDS = ""
    """
    Keywords indicating compatibility.
    Existence of the keyword indicates that the mod is stable on that platform.

    a ~ in front of the keyword indicates that the mod is unstable on that platform
    no keyword indicates that the mod is untested on that platform
    a - in front of the keyword indicates that the mod is known to not work on that platform

    E.g. A package that works on OpenMW but does not on tes3mp::

        KEYWORDS='openmw -tes3mp'
    """
    LICENSE = ""
    """
    One or more licenses used by the package.

    A list of licenses can be found in the licenses directory of the repository.
    """
    NAME = ""
    """
    Descriptive package name.

    The package name used for identification is the name used in the filename, however
    this name is included when searching for packages.
    """
    DESC = ""
    """
    A short description of the package.

    Is may (depending on options provided) be used in searches.
    Note that a longer description can be provided in metadata.yaml.
    """
    HOMEPAGE = ""
    """
    The URL of the package's homepage(s).

    Used for descriptive purposes and included in search results.
    """
    RESTRICT = ""
    """
    Lists features which should be disabled for this package

    The following two options are supported:

    * mirror: The package's SRC_URI entries should not be mirrored, and mirrors
      should not be checked when fetching.
    * fetch: The packages's SRC_URI entries should not be fetched automatically,
      and the pkg_nofetch function should be invoked if a source cannot be found.
      This option implies mirror.

    Note that portmod also supports determining these automatically based on source
    URIs and licenses, so it is no longer necessary to set them explicitly. mirror
    is restricted for licenses which are not in the REDISTRIBUTABLE license group
    (see license_groups.yaml), and fetch is restricted for files which are not
    redistributable (according to license) and do not have a scheme in their
    SRC_URI (i.e. just a filename, no https://domain.tld etc.).
    """
    PROPERTIES = ""
    """
    A white-space-delimited list of additional properties of the given pybuild to
    enable special behaviour.

    Possible values are given below:

    * ``live``: Indicates that the pybuild doesn't have a specific version (e.g. if
        installing from a git repository branch but not using a specific commit).
        Live pybuilds should have an empty KEYWORDS list, as stability testing is
        not meaningful if the upstream source is changing. Live packages must override
        :py:func:`Pybuild2.can_update_live`.
    * ``local``: Only used internally to refer to Local mods with generated metadata
    """
    TEXTURE_SIZES = ""
    """
    A field declaring the texture size options that the package supports.

    If only one texture size option is available, this field need not be included.
    Texture sizes should be numbers representing the size of the texture in pixels.
    Given that textures are usually two-dimensional, the convention is to use:
    :math:`\\sqrt{ l \\cdot w}`

    E.g.::

        TEXTURE_SIZES = "1024 2048"

    This is a special type of USE_EXPAND variable, as use flags are created for its
    values in the form texture_size_SIZE (in the above example texture_size_1024 and
    texture_size_2048).

    These use flags can (and should) be used in the pybuild to enable sources and
    InstallDirs conditionally depending on whether or not the texture size was selected.
    Exactly one of these use flags will be enabled when the mod is installed depending
    on the value of the TEXTURE_SIZE variable in the user's portmod.cfg.

    Not included in the PMS
    """
    REQUIRED_USE = ""
    """
    An expression indicating valid combinations of use flags.

    Consists of a string containing sub-expressions of the form given below.
    Note that the brackets can contain arbitrary nested expressions of this form, and
    are not limited to what is shown in the examples below.

    ===================================================   ============================
    Behaviour                                             Expression
    ===================================================   ============================
    flag must be enabled                                  ``flag``
    flag must not be enabled                              ``!flag``
    If flag1 enabled then flag2 enabled                   ``flag1? ( flag2 )``
    If flag1 disabled then flag2 enabled                  ``!flag1? ( flag2 )``
    If flag1 disabled then flag2 disabled	          ``!flag1? ( !flag2 )``
    Must enable any one or more (inclusive or)            ``|| ( flag1 flag2 flag3 )``
    Must enable exactly one but not more (exclusive or)   ``^^ ( flag1 flag2 flag3 )``
    May enable at most one                                ``?? ( flag1 flag2 flag3 )``
    ===================================================   ============================
    """
    SRC_URI = ""
    """
    A List of sources to be fetched.

    If source files should be renamed, this can be done with the arrow operator as
    shown in the example below.

    Sources can be wrapped in use-conditional expressions to prevent certain sources
    from being downloaded unless certain use flags are set or unset.

    E.g.::

        SRC_URI=\"\"\"
            http://mw.modhistory.com/file.php?id=9321 -> FileName-1.0.zip
            flag? ( https://cdn.bethsoft.com/elderscrolls/morrowind/other/masterindex.zip )
        \"\"\"

    Note that if you are renaming files, they should correspond to the original
    filename as best possible, but should also contain version information of some sort
    to prevent conflicts with other sources from the same package. That is, if the
    package is updated, we do not want the updated source name to be the same as a
    previous source name, even if the source name did not change upstream.
    """
    PATCHES = ""
    """
    A list of patch files stored within the package's files directory in the repository

    Note that unlike as specified in the PMS, their paths must be relative to the
    files directory.

    See :func:`~apply_patch` for details on the supported patch format.
    """
    S = None
    """
    Specifies the default working directory for src_* functions.

    The default value (if S is None) is the name (minus extension) of the first source
    in SRC_URI (after use-conditionals have been evaluated).
    If this path does not exist, the working directory falls back to WORKDIR.

    This is also used to determine the base source path used for installing a InstallDir
    in the default src_install if S is not defined on the InstallDir.
    """
    DOCS: List[str] = [
        "README*",
        "readme*",
        "ReadMe*",
        "ChangeLog",
        "CHANGELOG*",
        "AUTHORS*",
        "NEWS*",
        "TODO*",
        "CHANGES*",
        "THANKS*",
        "BUGS*",
        "FAQ*",
        "CREDITS*",
        "Doc/*",
        "doc/*",
        "docs/*",
        "Docs/*",
    ]
    """
    A list of documentation patterns for the default ``src_install`` to install using :py:attr:`Pybuild2.dodoc`
    """
    _PYBUILD_VER = 2

    def src_prepare(self):
        if self.PATCHES:
            enabled = self.get_use()
            for patch in use_reduce(self.PATCHES, enabled, flat=True):
                path = os.path.join(self.FILESDIR, patch)
                apply_patch(path)

    def src_install(self):
        for pattern in self.DOCS:
            self.dodoc(pattern)

    def pkg_postinst(self):
        pass

    def pkg_prerm(self):
        pass

    def pkg_pretend(self):
        pass

    def unpack(self, archives: Union[str, Iterable[Union[Source, str]]]):
        """
        Unpacks the given archive into the workdir

        Uses :py:func:`shutil.unpack_archive` as its backend, supporting the following
        archive formats:
        - .zip
        - .tar
        - .tar.xz / .txz
        - .tar.bz2 / .tbz2
        - .tar.xz / .txz

        args:
            archives: The list of archives to be unpacked
        """
        if isinstance(archives, str):
            archives = [archives]

        for archive in archives:
            if isinstance(archive, str):
                name = archive
                path = archive
            else:
                path = archive.path
                name = archive.name

            info(">>> " + l10n("pkg-unpacking-source", archive=name))
            archive_name, ext = os.path.splitext(os.path.basename(name))
            # Hacky way to handle tar.etc having multiple extensions
            if archive_name.endswith("tar"):
                archive_name, _ = os.path.splitext(archive_name)
            outdir = os.path.join(self.WORKDIR, archive_name)
            os.makedirs(outdir)
            shutil.unpack_archive(path, outdir)

    def src_unpack(self):
        self.unpack(self.A)

    def can_update_live(self) -> bool:
        """
        Indicates whether or not a live package can be updated.

        The default implementation just returns False. If the package has ``live`` in its
        ```PROPERTIES``, it must implement this method.

        returns:
            If the package has ``PROPERTIES="live"`` and can be updated, should return True
            Otherwise, should return False
        """
        return False

    @staticmethod
    def execute(
        command: str, pipe_output: bool = False, pipe_error: bool = False
    ) -> Optional[str]:
        """
        Allows execution of arbitrary commands at runtime.
        Command is sandboxed with filesystem and network access depending on
        the context in which it is called

        args:
            command: Command to be executed
            pipe_output: If true, returns the output of the command
        """
        raise Exception("execute was called from an invalid context")

    def warn(self, string: str):
        """
        Displays warning message both immediately, and in the summary after all
        transactions have been completed

        args:
            string: String to display
        """
        directory = os.path.realpath(os.path.join(warnings_dir(), self.CATEGORY))
        os.makedirs(directory, exist_ok=True)
        with open(os.path.join(directory, self.ATOM.PF), "a+") as file:
            print(string, file=file)
        warning(string)

    def info(self, string: str):
        """
        Displays info message both immediately, and in the summary after all
        transactions have been completed

        args:
            string: String to display
        """
        directory = os.path.realpath(os.path.join(messages_dir(), self.CATEGORY))
        os.makedirs(directory, exist_ok=True)
        with open(os.path.join(directory, self.ATOM.PF), "a+") as file:
            print(string, file=file)
        info(string)

    def get_installed_env(self):
        """Returns a dictionary containing installed object values"""
        return get_installed_env(self)

    def dodoc(self, pattern: str):
        """
        Installs documentation matching the given pattern into the image directory
        (``Self.D``)

        args:
            pattern: A pattern which can include glob-style wildcards as implemented by
                     :py:mod:`glob`.
        """
        if not self.D:
            raise Exception("dodoc must be executed within src_install")
        dest = os.environ.get("DOC_DEST") or "doc"

        os.makedirs(os.path.join(self.D, dest, self.PN), exist_ok=True)

        for file in iglob(pattern):
            info(
                l10n(
                    "installing-doc-into",
                    src=os.path.relpath(file, self.WORKDIR),
                    dest=dest,
                )
            )
            # Handle multiple files of the same name.
            #
            # Such files will cause errors on Windows, however it is also a logic error since
            # different pieces of documentation with the same name are probably from
            # different parts of the package (e.g. main mod and a patch) and should
            # probably all be installed side by side.
            #
            # Instead, rename them so they have unique names.
            if os.environ.get("CASE_INSENSITIVE_FILES", False):
                path = os.path.join(dest, self.PN, os.path.basename(file))
                ci_path = ci_exists(path, prefix=self.D)
                # Skip file if identical to one which already exists
                if ci_path and cmp(file, ci_path, shallow=False):
                    continue
                if ci_path:
                    dest_path = make_unique_filename(
                        os.path.join(self.D, path), case_insensitive=True
                    )
                else:
                    dest_path = os.path.join(self.D, dest, self.PN)
            else:
                dest_path = os.path.join(self.D, dest, self.PN, os.path.basename(file))
                if os.path.exists(dest_path):
                    dest_path = make_unique_filename(os.path.join(self.D, path))

            try:
                shutil.move(file, dest_path)
            except OSError:
                shutil.copy2(file, dest_path)

    def validate(self):
        """
        *(Since Portmod 2.4)* inquisitor will call this function when scanning repositories.

        Only code that would be valid in the package global scope (see :ref:`sandbox`) may be
        used.

        This is designed to allow basic structural checks without hindering package loading.

        It differs from :py:func:`Pybuild2.pkg_pretend` in that it is meant for static checks,
        not runtime checks.
        """


class Pybuild1(Pybuild2):
    """
    Legacy class. Superseded by Pybuild2

    .. warning::
        This class has been deprecated as of Portmod 2.4.
        It will be removed in Portmod 3.0
    """

    DATA_OVERRIDES = ""
    """
    A use-reduce-able list of atoms indicating packages whose data directories should
    come before the data directories of this package when sorting data directories.

    They do not need to be dependencies. Blockers (atoms beginning with !!) can be used
    to specify underrides, and use dependencies (e.g. the [bar] in foo[bar]) can be
    used to conditionally override based on the target atom's flag configuration.

    Not included in PMS
    """
    INSTALL_DIRS: List[InstallDir] = []
    """
    The INSTALL_DIRS variable consists of a python list of InstallDir objects.

    E.g.::

        INSTALL_DIRS=[
            InstallDir(
                'Morrowind/Data Files',
                REQUIRED_USE='use use ...',
                DESTPATH='.',
                PLUGINS=[File('Plugin Name',
                    REQUIRED_USE='use use ...', satisfied
                )],
                ARCHIVES=[File('Archive Name')],
                S='Source Name Without Extension',
            )
        ]

    Not included in PMS
    """
    REBUILD_FILES: List[str] = []
    """
    Files in the VFS which, if changed, should cause this package to be rebuilt

    Can include glob-style patterns using the *, ? and [] operators.
    See https://docs.python.org/3/library/fnmatch.html.
    Unlike normal fnmatch parsing, wildcards (*) will not match accross
    path separators.

    This field can be modified during installation, and will only be used after the
    package has been installed.
    """
    TIER = "a"
    """
    The Tier of a package represents the position of its data directories and plugins
    in the virtual file system.

    This is used to group packages in such a way to avoid having to individually
    specify overrides whenever possible.

    The value is either in the range [0-9] or [a-z].

    Default value: 'a'

    Tier 0 represents top-level mods such as morrowind
    Tier 1 is for mods that replace or modify top-level mods. E.g. texture and mesh replacers.
    Tier 2 is for large mods that are designed to be built on top of by other mods, such as Tamriel Data
    Tier a is for all other mods.
    Tier z is for mods that should be installed or loaded last. E.g. omwllf
    The remaining tiers are reserved in case the tier system needs to be expanded

    Not included in PMS
    """
    _PYBUILD_VER = 1

    def __init__(self):
        super().__init__()
        if type(self.TIER) is int:
            self.TIER = str(self.TIER)
        elif type(self.TIER) is not str:
            raise TypeError("TIER must be a integer or string containing 0-9 or z")

    def _get_install_dir_dest_specific(self, install_dir: InstallDir) -> str:
        install_dir_dest = _get_install_dir_dest(self)

        if install_dir.RENAME is None:
            return os.path.normpath(
                os.path.join(self.D, install_dir_dest, install_dir.PATCHDIR)
            )
        else:
            return os.path.normpath(
                os.path.join(
                    self.D,
                    install_dir_dest,
                    install_dir.PATCHDIR,
                    install_dir.RENAME,
                )
            )

    def get_files(self, typ: str) -> Generator[Tuple[InstallDir, File], None, None]:
        """
        Returns all enabled files and their directories
        """
        yield from get_files(self, typ)

    def unpack(self, archives: Union[str, Iterable[Union[Source, str]]]):
        """
        Unpacks the given archive into the workdir

        Uses `patool <http://wummel.github.io/patool/>`_  as its backend.

        args:
            archives: The list of archives to be unpacked
        """
        # Slow import
        import patoolib

        if isinstance(archives, str):
            archives = [archives]

        for archive in archives:
            if isinstance(archive, str):
                name = archive
                path = archive
            else:
                path = archive.path
                name = archive.name

            info(">>> " + l10n("pkg-unpacking-source", archive=name))
            archive_name, ext = os.path.splitext(os.path.basename(name))
            # Hacky way to handle tar.etc having multiple extensions
            if archive_name.endswith("tar"):
                archive_name, _ = os.path.splitext(archive_name)
            outdir = os.path.join(self.WORKDIR, archive_name)
            os.makedirs(outdir)
            if logging.root.level >= logging.WARN:
                verbosity = -1
            else:
                verbosity = 0
            patoolib.extract_archive(
                path, outdir=outdir, interactive=False, verbosity=verbosity
            )

    def _install_directory(
        self, source: str, install_dir: InstallDir, to_install: Set[str]
    ):
        case_insensitive = os.environ.get("CASE_INSENSITIVE_FILES", False)

        dest = self._get_install_dir_dest_specific(install_dir)

        os.makedirs(dest, exist_ok=True)
        blacklist_entries = install_dir.BLACKLIST or []

        for file in install_dir.get_files():
            # ignore files which will not be used
            if not check_required_use(
                file.REQUIRED_USE, self.get_use(), self.valid_use
            ):
                blacklist_entries.append(file.NAME)

        blacklist = fnmatch_list_to_re(blacklist_entries)
        if install_dir.WHITELIST is None:
            whitelist = None
        else:
            whitelist = fnmatch_list_to_re(install_dir.WHITELIST)

        def get_listfn(filter_re: Pattern, polarity: bool):
            def fn(directory: str, contents: Iterable[str]):
                paths = []
                basedir = os.path.relpath(directory, source)
                for file in contents:
                    path = os.path.normpath(os.path.join(basedir, file))
                    paths.append(path)

                if polarity:
                    return {
                        file
                        for path, file in zip(paths, contents)
                        if filter_re.match(path)
                        and not os.path.isdir(os.path.join(directory, file))
                    }
                else:
                    return {
                        file
                        for path, file in zip(paths, contents)
                        if not filter_re.match(path)
                        and not os.path.isdir(os.path.join(directory, file))
                    }

            return fn

        # Function to ingore additional paths, given an existing ignore function
        def ignore_more(
            ignorefn, to_ignore: List[str]
        ) -> Callable[[str, List[str]], Set[str]]:
            def fn(directory: str, contents: Iterable[str]):
                results = ignorefn(directory, contents)
                for name in contents:
                    if any(
                        os.path.normpath(os.path.join(directory, name))
                        == os.path.normpath(path)
                        for path in to_ignore
                    ):
                        results.add(name)
                return results

            return fn

        # Determine if any other InstallDirs are inside this one and add them to the blacklist
        to_ignore = []
        for other_path in to_install:
            if other_path != source and is_parent(
                os.path.abspath(other_path), os.path.abspath(source)
            ):
                to_ignore.append(other_path)

        if os.path.islink(source):
            linkto = os.readlink(source)
            if os.path.exists(dest):
                if os.path.islink(dest):
                    os.remove(dest)
                else:
                    os.rmdir(dest)
            os.symlink(linkto, dest, True)
        elif whitelist is not None:
            ignore = get_listfn(whitelist, False)
            if to_ignore:
                ignore = ignore_more(ignore, to_ignore)
            patch_dir(source, dest, ignore=ignore, case_sensitive=not case_insensitive)
        elif blacklist_entries:
            ignore = get_listfn(blacklist, True)
            if to_ignore:
                ignore = ignore_more(ignore, to_ignore)
            patch_dir(source, dest, ignore=ignore, case_sensitive=not case_insensitive)
        else:
            ignore = None
            if to_ignore:
                ignore = ignore_more(lambda _d, _c: set(), to_ignore)
            patch_dir(source, dest, case_sensitive=not case_insensitive, ignore=ignore)

    def src_install(self):
        """
        The src_install function installs the package’s content to a directory specified
        in :py:attr:`Pybuild2.D`.

        The initial working directory is :py:attr:`Pybuild2.S`, falling back to
        :py:attr:`Pybuild.WORKDIR` if the directory does not exist.

        The default implementation used when the package lacks the ``src_install`` function
        moves each InstallDir in :py:attr:`Pybuild1.INSTALL_DIRS` which is not hidden due
        to an unsatisfied ``REQUIRED_USE`` into :py:attr:`Pybuild2.D`.
        """
        to_install: List[Tuple[str, InstallDir, bool]] = []
        sources = set()
        for install_dir in self.INSTALL_DIRS:
            source_dir = install_dir.S or cast(str, self.S)
            source = os.path.normpath(
                os.path.join(self.WORKDIR, source_dir, install_dir.PATH)
            )
            sources.add(source)
            if check_required_use(
                install_dir.REQUIRED_USE, self.get_use(), self.valid_use
            ):
                # self.S will be set in package.py via the PhaseState
                to_install.append((source, install_dir, True))
            else:
                to_install.append((source, install_dir, False))

        for (source, install_dir, enabled) in to_install:
            source_dir_path = os.path.join(
                install_dir.S or cast(str, self.S), install_dir.PATH
            )

            if enabled:
                info(
                    l10n(
                        "installing-directory-into",
                        dir=magenta(source_dir_path),
                        dest=magenta(self._get_install_dir_dest_specific(install_dir)),
                    )
                )
                for doc_path in set(install_dir.DOC) | set(self.DOCS):
                    self.dodoc(os.path.join(self.WORKDIR, source_dir_path, doc_path))
                self._install_directory(source, install_dir, sources)
            else:
                print(
                    l10n(
                        "skipping-directory",
                        dir=magenta(source_dir_path),
                        req=blue(install_dir.REQUIRED_USE),
                    )
                )
